/*
 *  TOPPERS BASE PLATFORM MIDDLEWARE
 * 
 *  Copyright (C) 2017-2024 by TOPPERS PROJECT
 *                             Educational Working Group.
 * 
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 * 
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 * 
 *  @(#) $Id: protocal.c 698 2024-03-20 18:32:25Z roi $
 */
/*
 *  BLE MOTOR制御 Middleware
 */

#include <stdint.h>
#include <string.h>
#include "protocol.h"
#include <stdio.h>

/*
 *  バージョン情報
 */
#define	BLEREM_PRVER	0x0001		/* BLE REMOTEのバージョン番号 */

/*
 *  コマンドデスパッチ用の構造体定義
 */
typedef struct _COMMAND_t {
	const char   *command;			/* コマンド文 */
	int          (*func)(char *res, char *cmd, uint16_t *plen);	/* 実行関数 */
} COMMAND_t;

static int info_func(char *res, char *cmd, uint16_t *plen);
static int gdev_func(char *res, char *cmd, uint16_t *plen);
static int ver_func(char *res, char *cmd, uint16_t *plen);
static int gmtr_func(char *res, char *cmd, uint16_t *plen);
static int gsys_func(char *res, char *cmd, uint16_t *plen);
static int sinf_func(char *res, char *cmd, uint16_t *plen);
static int sdev_func(char *res, char *cmd, uint16_t *plen);
static int smtr_func(char *res, char *cmd, uint16_t *plen);
static int suss_func(char *res, char *cmd, uint16_t *plen);

/*
 *  デバイスコマンドテーブル
 */
static const COMMAND_t command_info[] = {
	{"GI",		info_func},
	{"GD",		gdev_func},
	{"GV",		ver_func},
	{"GM",		gmtr_func},
	{"GS",		gsys_func},
	{"SI",		sinf_func},
	{"SD",		sdev_func},
	{"SM",		smtr_func},
	{"SU",		suss_func}
};

#define NUM_COMMAND  (sizeof(command_info)/sizeof(COMMAND_t))

static int8_t  t_duty[NUM_MOTORS][2];
static int8_t  s_duty[NUM_MOTORS][2];

static char *
int2char(char *d, int len)
{
	int  i;
	uint8_t buf[16];

	if(len == 0){
		*d++ = '0';
		return d;
	}
	for(i = 0 ; i < 16 && len != 0 ; i++){
		buf[i] = (len % 10) + '0';
		len /= 10;
	}
	while(i > 0){
		*d++ = buf[--i];
	}
	return d;
}

static int
a2i(char *str, char **p)
{
	int sign = 1;
	int num = 0;

	if(*str == 'S' || *str == 's'){
		*p = str+2;
		return 127;
	}
	if(*str == '-'){
		sign = -1;
		str++;
	}
	while(*str >= '0' && *str <= '9'){
		num = num * 10 + *str++ - '0';
	}
	*p = str+1;
	return num * sign;
}

/*
 *  EEPROMにデータを書込む
 */
uint32_t
eeprom_write(uint16_t addr, uint8_t const *pdata, uint8_t size)
{
	uint8_t buffer[EEPROM_ADDRESS_LEN_BYTES + EEPROM_WRITE_MAX_BYTES]; /* Addr + data */
    uint32_t err_code;

	if(size > EEPROM_WRITE_MAX_BYTES)
		return EEPROM_INVALID_LENGTH;
	if((addr / EEPROM_WRITE_MAX_BYTES) != ((addr + size - 1) / EEPROM_WRITE_MAX_BYTES))
		return EEPROM_INVALID_ADDR;

	buffer[0] = addr >> 8;
	buffer[1] = addr & 0xff;
	memcpy(buffer + EEPROM_ADDRESS_LEN_BYTES, pdata, size);
	err_code = twi_write(EEPROM_ADDR, buffer, size + EEPROM_ADDRESS_LEN_BYTES, 0);
	return err_code;
}

/*
 *  EEPROMからデータを読み込む
 */
uint32_t
eeprom_read(uint16_t addr, uint8_t *pdata, uint16_t size)
{
	uint32_t err_code;
	uint8_t addr8[2];

	addr8[0] = addr >> 8;
	addr8[1] = addr & 0xFF;
	err_code = twi_write(EEPROM_ADDR, addr8, EEPROM_ADDRESS_LEN_BYTES, 1);
	if(err_code != 0)
		return err_code;
	err_code = twi_read(EEPROM_ADDR, pdata, size);
    return err_code;
}


/*
 *  モータ初期化処理
 */
void
motor_init(void)
{
	uint32_t mid;

	for(mid = 0 ; mid < NUM_MOTORS ; mid++){
		t_duty[mid][0] = t_duty[mid][1] = 0;
		s_duty[mid][0] = s_duty[mid][1] = 0;
		motor_stby_set(mid, 0);
		motor_duty_set(mid, s_duty[mid][0], s_duty[mid][1]);
	}
}


/*
 *  モータイベント処理
 */
uint32_t
motor_event(void)
{
	EEPROM_t *pf = get_eeprom();
	uint32_t next_time = 0;
	int8_t   ss_duty[2];
	uint32_t mid;

	for(mid = 0 ; mid < NUM_MOTORS ; mid++){
		if(s_duty[mid][0] != t_duty[mid][0] || s_duty[mid][1] != t_duty[mid][1]){
			uint8_t id  = pf->t_mselect ^ mid;
			uint8_t pno = pf->t_m_pulse[mid];
			if(s_duty[mid][0] > 0 && s_duty[mid][0] <= 100 && t_duty[mid][1] > 0){
				ss_duty[0] = 0;
				ss_duty[1] = s_duty[mid][1];
				next_time = pf->t_m_time;
			}
			else if(s_duty[mid][1] > 0 && s_duty[mid][1] <= 100 && t_duty[mid][0] > 0){
				ss_duty[0] = s_duty[mid][0];
				ss_duty[1] = 0;
				next_time = pf->t_m_time;
			}
			else{
				ss_duty[0] = s_duty[mid][0];
				ss_duty[1] = s_duty[mid][1];
			}
			motor_stby_set(id, 1);
			motor_duty_set(id, ss_duty[0 ^ pno], ss_duty[1 ^ pno]);
			t_duty[mid][0] = ss_duty[0];
			t_duty[mid][1] = ss_duty[1];
			if(ss_duty[0] == 0 && ss_duty[1] == 0)
				motor_stby_set(id, 0);
		}
	}
	return next_time;
}


/*
 *  超音波センサーセンス開始
 */
int
uss_start(void)
{
#if defined(USE_I2C_ULTRASONIC_SENSOR)
    uint32_t err_code;
	uint8_t buffer[2];

	buffer[0] = US_SENSOR_START_CMD;
	err_code = twi_write(US_SENSOR_ADDR, buffer, 1, 0);
	if(err_code == 0)
		err_code = twi_read(US_SENSOR_ADDR, buffer, 1);
	if(err_code == 0)
		return buffer[0];
	else
#endif
		return -1;
}


/*
 *  超音波センサー結果取得
 */
int
uss_get_result(uint8_t uss_value)
{
#if defined(USE_I2C_ULTRASONIC_SENSOR)
    uint32_t err_code;
	uint8_t buffer[2], datah;
	int result = 0;

	if(uss_value != 0){
		buffer[0] = US_SENSOR_DATAH_CMD;
		err_code = twi_write(US_SENSOR_ADDR, buffer, 1, 0);
		buffer[0] = 0;
		if(err_code == 0)
			err_code = twi_read(US_SENSOR_ADDR, buffer, 1);
		datah = buffer[0];

		buffer[0] = US_SENSOR_DATAL_CMD;
		if(err_code == 0)
			err_code = twi_write(US_SENSOR_ADDR, buffer, 1, 0);
		buffer[0] = 0;
		if(err_code == 0)
			err_code = twi_read(US_SENSOR_ADDR, buffer, 1);
		result = datah*256 + buffer[0];
	}
	return result;
#else
	return -1;
#endif
}


/*
 *  超音波センサーのリポート作成
 */
int
uss_make_event(char *res, uint32_t uss_value1, uint32_t uss_value2, uint16_t *plen)
{
	char *p = res;

	*p++ = 'R';
	*p++ = 'U';
	*p++ = ':';
	*p++ = '0'+uss_value1;
	*p++ = ',';
	p = int2char(p, uss_value2);
	*p = 0;
	*plen = p - res;
	return 0;
}


/*
 *  コマンド解析と実行
 */
int
dispatch(char *res, char *cmd, uint16_t *plen)
{
	const COMMAND_t *pc;
	char  *p = &cmd[2];
	uint16_t len;
	int i;

	if(cmd == NULL || plen == NULL)
		return EVENT_IDOL;
	len = *plen;
	if(len < 2)
		return EVENT_IDOL;
	if(cmd[len-1] < 0x20)
		len--;
	if(cmd[len-1] < 0x20)
		len--;
	if(len < 2)
		return EVENT_IDOL;

	cmd[len] = 0;
	*plen = len;
	for(i = 0, pc = command_info ; i < NUM_COMMAND ; i++, pc++){
		if(cmd[0] >= 'a' && cmd[0] < 'z')
			cmd[0] -= 0x20;
		if(cmd[1] >= 'a' && cmd[1] < 'z')
			cmd[1] -= 0x20;
		if(pc->command[0] == cmd[0] && pc->command[1] == cmd[1]){
			*plen -= 2;
			if(*plen > 0 && *p == ':'){
				p++;
				*plen -= 1;
			}
			return pc->func(res, p, plen);
		}
	}
	return EVENT_IDOL;
}

/*
 *  レスポンス：インフォメーション
 */
static int
info_func(char *res, char *cmd, uint16_t *plen)
{
	EEPROM_t *pf = get_eeprom();
	char *p = res;

	*p++ = 'R';
	*p++ = 'I';
	*p++ = ':';
	p = int2char(p, pf->t_mselect);
	*p++ = ',';
	p = int2char(p, pf->t_m_pulse[MOTOR1_ID]);
	*p++ = ',';
	p = int2char(p, pf->t_m_pulse[MOTOR2_ID]);
	*p++ = ',';
	p = int2char(p, pf->t_m_bias[MOTOR1_ID]);
	*p++ = ',';
	p = int2char(p, pf->t_m_bias[MOTOR2_ID]);
	*p = 0;
	*plen = p - res;
	return EVENT_SEND_RESPONSE;
}

/*
 *  レスポンス：デバイス名
 */
static int
gdev_func(char *res, char *cmd, uint16_t *plen)
{
	EEPROM_t *pf = get_eeprom();
	char *p = res;
	uint32_t i;

	*p++ = 'R';
	*p++ = 'D';
	*p++ = ':';
	for(i = 0 ; i < 32 && pf->devicename[i] != 0 ; i++){
		*p++ = pf->devicename[i];
	}
	*p = 0;
	*plen = p - res;
	return EVENT_SEND_RESPONSE;
}

/*
 *  レスポンス：バージョン
 */
static int
ver_func(char *res, char *cmd, uint16_t *plen)
{
	char *p = res;

	*p++ = 'R';
	*p++ = 'V';
	*p++ = ':';
	p = int2char(p, ((BLEREM_PRVER>>12) & 0xf));
	*p++ = '.';
	p = int2char(p, ((BLEREM_PRVER>>4) & 0xff));
	*p++ = '.';
	p = int2char(p, (BLEREM_PRVER & 0xf));
	*p = 0;
	*plen = p - res;
	return EVENT_SEND_RESPONSE;
}

/*
 *  モータ設定取得
 */
static int
gmtr_func(char *res, char *cmd, uint16_t *plen)
{
	char *p = res;
	uint32_t mid;
	int8_t   val = 0;

	*p++ = 'R';
	*p++ = 'M';
	*p++ = ':';
	for(mid = 0 ; mid < NUM_MOTORS ; mid++){
		if(t_duty[mid][0] == 100 && t_duty[mid][1] == 100)
			*p++ = 'S';
		else{
			if(t_duty[mid][0] == 0 && t_duty[mid][1] == 0)
				val = 0;
			else if(t_duty[mid][0] > 0)
				val = t_duty[mid][0];
			else{
				*p++ = '-';
				val = t_duty[mid][1];
			}
			p = int2char(p, val);
		}
		if((mid+1) < NUM_MOTORS)
			*p++ = ',';
	}
	*p = 0;
	*plen = p - res;
	return EVENT_SEND_RESPONSE;
}

/*
 *  システム情報取得
 */
static int
gsys_func(char *res, char *cmd, uint16_t *plen)
{
	char *p = res;

	*p++ = 'R';
	*p++ = 'S';
	*p++ = ':';
	p = int2char(p, get_systime());
	*p++ = ',';
	p = int2char(p, get_batterylevel());
	*p++ = ',';
	p = int2char(p, get_syserror());
	*p = 0;
	*plen = p - res;
	return EVENT_SEND_RESPONSE;
}

/*
 *  モータ実行設定
 */
static int
sinf_func(char *res, char *cmd, uint16_t *plen)
{
	EEPROM_t *pf = get_eeprom();
	uint8_t len = *plen;
	uint32_t i = 0, val;
	char *p = cmd;
	char **pp = &p;

	if(i < len){
		if(cmd[0] != ','){
			val = a2i(cmd, pp);
			if(val == 0 || val == 1)
				pf->t_mselect = val;
			for(; i < len ; i++){
				res[i] = *cmd++;
				if(res[i] == ',')
					break;
			}
		}
		else
			res[i++] = *cmd++;
	}
	if(i < len){
		if(cmd[0] != ','){
			val = a2i(cmd, pp);
			if(val == 0 || val == 1)
				pf->t_m_pulse[MOTOR1_ID] = val;
			for(; i < len ; i++){
				res[i] = *cmd++;
				if(res[i] == ',')
					break;
			}
		}
		else
			res[i++] = *cmd++;
	}
	if(i < len){
		if(cmd[0] != ','){
			val = a2i(cmd, pp);
			if(val == 0 || val == 1)
				pf->t_m_pulse[MOTOR2_ID] = val;
			for(; i < len ; i++){
				res[i] = *cmd++;
				if(res[i] == ',')
					break;
			}
		}
		else
			res[i++] = *cmd++;
	}
	if(i < len){
		if(cmd[0] != ','){
			val = a2i(cmd, pp);
			if(val <= 10000 && val >= 100)
				pf->t_m_bias[MOTOR1_ID] = val;
			for(; i < len ; i++){
				res[i] = *cmd++;
				if(res[i] == ',')
					break;
			}
		}
		else
			res[i++] = *cmd++;
	}
	if(i < len){
		if(cmd[0] != ','){
			val = a2i(cmd, pp);
			if(val <= 10000 && val >= 100)
				pf->t_m_bias[MOTOR2_ID] = val;
			for(; i < len && res[i] != ',' ; i++)
				res[i] = *cmd++;
		}
		for(; i < len ; i++){
			res[i] = *cmd++;
		}
	}
	res[i] = 0;
	*plen = i;
	return EVENT_WRITE_FLASH;
}

/*
 *  デバイス名設定
 */
static int
sdev_func(char *res, char *cmd, uint16_t *plen)
{
	EEPROM_t *pf = get_eeprom();
	uint8_t len = *plen;
	uint32_t i;

	if(len < 6)
		return EVENT_IDOL;
	if(cmd[0] != 'B' || cmd[1] != 'L' ||
			cmd[2] != 'E' || cmd[3] != '_')
		return EVENT_IDOL;
	for(i = 0 ; i < 32 && i < len ; i++){
		pf->devicename[i] = res[i] = *cmd++;
	}
	pf->devicename[i] = res[i] = 0;
	*plen = i;
	return EVENT_WRITE_FLASH;
}

/*
 *  モータ設定
 */
static int
smtr_func(char *res, char *cmd, uint16_t *plen)
{
	EEPROM_t *pf = get_eeprom();
	uint8_t  len = *plen;
	int32_t  val;
	uint32_t mid;
	char *p = cmd;
	char **pp = &p;

	memcpy(s_duty, t_duty, NUM_MOTORS*2);
	for(mid = 0 ; mid < NUM_MOTORS ; mid++){
		if((len - (*pp - cmd)) > 0){
			if(p[0] == ',' || p[0] == ' ')
				p++;
			else{
				val = a2i(p, pp);
				if(val == 127)
					s_duty[mid][0] = s_duty[mid][1] = 100;
				else if(val <= 100 && val >= -100){
					val = (val * pf->t_m_bias[mid]) / 10000;
					if(val == 0)
						s_duty[mid][0] = s_duty[mid][1] = 0;
					else if(val > 0){
						s_duty[mid][0] = val;
						s_duty[mid][1] = 0;
					}
					else{
						s_duty[mid][0] = 0;
						s_duty[mid][1] = -val;
					}
				}
				p = *pp;
			}
		}
	}
	return EVENT_MOTOR_EVENT;
}

/*
 *  超音波センサー(mi-02)センス値取得
 */
static int
suss_func(char *res, char *cmd, uint16_t *plen)
{
	EEPROM_t *pf = get_eeprom();
	int32_t  val1 = 0;
	int32_t  val2;
	char *p = cmd;
	char **pp = &p;

	val1 = a2i(p, pp);
	val2 = a2i(p, pp);
	if(val1 < pf->u_s_interval)
		val1 = 0;
	if(val2 < pf->u_s_time)
		val2 = pf->u_s_time;
	uss_event_request(val1, val2);
	*res = 0;
	*plen = 0;
	return EVENT_IDOL;
}

